import re
from typing import Callable, Iterable, List, Optional, Tuple

from bodhilib import BaseSplitter, Document, Node


class TextSplitter(BaseSplitter):
    """Splitter splits a :class:`~bodhilib.Document` into :class:`~bodhilib.Node`."""

    def __init__(
        self,
        *,
        max_len: int = 512,
        min_len: int = 128,
        overlap: int = 16,
        eos_patterns: Optional[List[str]] = None,
        eow_patterns: Optional[List[str]] = None,
    ) -> None:
        r"""Initializing splitter to split text based on sentence and word splits.

        Args:
            max_len (Optional[int]): Maximum number of words in split text. Defaults to 512.
            min_len (Optional[int]): Minimum number of words in split text. Defaults to 128.
            overlap (Optional[int]): Number of words to overlap between splits. Defaults to 16.
            eos_patterns (Optional[List[str]]): List of patterns to split sentences.
                The patterns should be regex. E.g. `[r"\n", r"\."]`.
                Defaults to `[r"\.", r"\?", r"\!", r"\n"]`.
            eow_patterns (Optional[List[str]]): List of patterns to split words.
                The patterns should be regex. E.g. `[r"\s", r"\-"]`.
                Defaults to `[r"\s", r"\-", r"\:", r"\.", r"\?", r"\!", r"\n"]`.
        """
        assert max_len > min_len, f"{max_len=} should be greater than {min_len=}"
        assert overlap < max_len, f"{overlap=} should be less than {max_len=}"
        assert overlap < min_len, f"{overlap=} should be less than {min_len=}"

        self.max_len = max_len
        self.min_len = min_len
        self.overlap = overlap

        if eos_patterns is None:
            eos_patterns = [r"\.", r"\?", r"\!", r"\n"]
        self.sentence_splitter = _build_sentence_splitter(eos_patterns)

        if eow_patterns is None:
            eow_patterns = [r"\s", r"-", r":", r"\.", r"\?", r"\!", r"\n"]
        self.word_splitter = _build_word_splitter(eow_patterns)

    def _split(self, docs: Iterable[Document]) -> List[Node]:
        nodes = []
        for doc in docs:
            current_words: List[str] = []
            sentences = self.sentence_splitter(doc.text)
            for sentence in sentences:
                words = self.word_splitter(sentence)
                # the sentence can be combined without exceeding max_len
                if len(current_words) + len(words) < self.max_len:
                    current_words += words
                    words = []
                    continue
                # the sentence cannot be combined without exceeding max_len
                new_nodes, new_current_words, new_words = self._build_nodes(doc, current_words, words)
                nodes.extend(new_nodes)
                assert new_words == [], f"{new_words=} should be empty"
                current_words = new_current_words
            if len(current_words) > self.overlap:
                node_text = "".join(current_words)
                node = Node(text=node_text, parent=doc)
                nodes.append(node)
        return nodes

    def _build_nodes(
        self, doc: Document, current_words: List[str], words: List[str]
    ) -> Tuple[List[Node], List[str], List[str]]:
        nodes = []
        while True:
            # start of sentence, take all the words
            if len(current_words) == 0:
                current_words = words
                words = []
                continue

            # the sentence can be combined with next sentence without exceeding max_len
            if len(words) != 0 and len(current_words) + len(words) < self.max_len:
                current_words += words
                words = []
                continue

            # current sentence has more words than max_len
            # take the max len words and build the node, pass the remaining words to next iteration
            if len(current_words) >= self.max_len:
                node_text = "".join(current_words[: self.max_len])
                remaining_words = current_words[self.max_len - self.overlap :]
                new_node = Node(text=node_text, parent=doc)
                nodes.append(new_node)
                current_words = remaining_words
                continue

            # if combined  with next sentence, the words will exceed max_len
            # and current words is more than min_len
            # so build the node with current words, pass the remaining words to next iteration
            if len(current_words) + len(words) > self.max_len and len(current_words) >= self.min_len:
                node_text = "".join(current_words)
                new_node = Node(text=node_text, parent=doc)
                nodes.append(new_node)
                remaining_words = current_words[-self.overlap :] if self.overlap != 0 else []
                current_words = remaining_words
                continue

            # if combined  with next sentence, the words will exceed max_len
            # and current words is less than min_len
            # so take as many words needed to reach min_len, build the node and return remaining words to next iteration
            if len(current_words) + len(words) > self.max_len and len(current_words) < self.min_len:
                all_words = current_words + words
                node_text = "".join(all_words[: self.min_len])
                new_node = Node(text=node_text, parent=doc)
                nodes.append(new_node)
                remaining_words = all_words[self.min_len - self.overlap :]
                current_words = remaining_words
                words = []
                continue

            # the combined sentence will be less than max_len
            # add the words and return
            current_words += words
            words = []
            break
        return nodes, current_words, words


def _build_sentence_splitter(eos_patterns: List[str]) -> Callable[[str], List[str]]:
    return _build_symbol_splitter(eos_patterns)


def _build_word_splitter(eow_patterns: List[str]) -> Callable[[str], List[str]]:
    return _build_symbol_splitter(eow_patterns)


def _build_symbol_splitter(symbols: List[str]) -> Callable[[str], List[str]]:
    # Construct the non-symbol pattern
    non_symbol_pattern = f"(?:(?!{'|'.join(symbols)}).)+"

    # Construct the symbol pattern
    symbol_pattern = "|".join(symbols)

    # Pattern to check if entire string consists of only symbols
    only_symbol_matcher = re.compile(f"^(?:{symbol_pattern})+$")

    # Modify the word pattern to capture leading symbols
    word_pattern = f"(?:{symbol_pattern})*{non_symbol_pattern}(?:{symbol_pattern})*"
    word_splitter = re.compile(word_pattern)

    def splitter(text: str) -> List[str]:
        if only_symbol_matcher.match(text):
            return [text]
        return word_splitter.findall(text)

    return splitter



from __future__ import annotations

import abc
from typing import Any, Dict, Iterable, Iterator, List, Optional, Type, TypeVar, Union, cast

from ._models import (
    Distance,
    Document,
    Embedding,
    Node,
    Prompt,
    PromptStream,
    PromptTemplate,
    SerializedInput,
    to_document_list,
    to_node_list,
    to_prompt_list,
)
from ._plugin import PluginManager, Service


# region prompt source
#######################################################################################################################
class PromptSource(abc.ABC):
    """Abstract base class for prompt sources.

    A prompt source provides a browsable/searchable interface for prompt templates.
    """

    @abc.abstractmethod
    def find(self, tags: Union[str, List[str]]) -> List[PromptTemplate]:
        """Find a prompt template for given tags.

        Args:
            tags (str | List[str]): list of tags to search for

        Returns:
            List[PromptTemplate]: list of prompt templates matching the tags
        """

    @abc.abstractmethod
    def list_all(self) -> List[PromptTemplate]:
        """List all prompt templates in the source.

        Returns:
            List[PromptTemplate]: list of all prompt templates in the source
        """


class BasePromptSource(PromptSource):
    """BasePromptSource provides a simpler method for implementing PromptSources.

    Class implementing :class:`~bodhilib.PromptSource` should extend this base class.
    In case the abstract method changes in :class:`~bodhilib.PromptSource`,
    this BaseClass tries to safely adapt new changes to old interface.
    """

    def find(self, tags: Union[str, List[str]]) -> List[PromptTemplate]:
        if isinstance(tags, str):
            tags = [tags]
        return self._find(tags)

    def list_all(self) -> List[PromptTemplate]:
        return self._list_all()

    @abc.abstractmethod
    def _find(self, tags: List[str]) -> List[PromptTemplate]:
        raise NotImplementedError(f"find method is not implemented by class {self.__class__.__name__}")

    @abc.abstractmethod
    def _list_all(self) -> List[PromptTemplate]:
        raise NotImplementedError(f"list_all method is not implemented by class {self.__class__.__name__}")


# endregion
# region data loader
#######################################################################################################################
class DataLoader(Iterable[Document], abc.ABC):
    """Abstract base class for data loaders.

    A data loader should inherit from this class and implement the abstract methods.
    """

    @abc.abstractmethod
    def add_resource(self, **kwargs: Dict[str, Any]) -> None:
        """Add a resource to the data loader."""

    @abc.abstractmethod
    def __iter__(self) -> Iterator[Document]:
        """Returns the document iterator.

        It is for the sub-class to ensure the `__iter__` method returns a new instance of iterator
        """

    def load(self) -> List[Document]:
        """Returns the document as list."""
        return list(self)


# endregion
# region splitter
#######################################################################################################################
class Splitter(abc.ABC):
    """Splitter defines abstract method to split longer text into shorter text.

    Splitter takes in longer text as a generic :data:`~bodhilib.TextLike`
    and splits them into shorter text and return as :class:`~bodhilib.Node`.
    The shorter text are then used to create embeddings.
    """

    @abc.abstractmethod
    def split(self, inputs: SerializedInput) -> List[Node]:
        """Split a :data:`~bodhilib.SerializedInput` into a list of :class:`~bodhilib.Node`.

        Args:
            inputs (:data:`~bodhilib.SerializedInput`): takes input as :data:`~bodhilib.SerializedInput`,
                a generic type that can be a :data:`~bodhilib.TextLike`, a list of :data:`~bodhilib.TextLike`,
                or a serialized dict of the object.

        Returns:
            List[:class:`~bodhilib.Node`]: a list of :class:`~bodhilib.Node` as result of the split
        """


class BaseSplitter(Splitter):
    """BaseSplitter provides a simpler method for implementing Splitters.

    BaseSplitter overrides the abstract Splitter method :func:`~bodhilib.Splitter.split`, massages the data
    and converts it to a list of :class:`~bodhilib.Document`
    and passes to implementing :func:`~bodhilib.BaseSplitter._split` method.
    """

    def split(self, inputs: SerializedInput) -> List[Node]:
        docs = to_document_list(inputs)
        return self._split(docs)

    @abc.abstractmethod
    def _split(self, docs: Iterable[Document]) -> List[Node]:
        """Split a list of :class:`~bodhilib.Document` into list of :class:`~bodhilib.Node`.

        The split method preserves the relationship and copies the metadata associated with Document to the Node.
        """


# endregion
# region embedder
#######################################################################################################################
class Embedder(abc.ABC):
    """Abstract base class for embedders.

    An embedder should inherit from this class and implement the abstract methods.
    """

    @abc.abstractmethod
    def embed(self, inputs: SerializedInput) -> List[Node]:
        """Embed a :data:`~bodhilib.SerializedInput` using the embedder service.

        Args:
            inputs (:data:`~bodhilib.SerializedInput`): takes input as :data:`~bodhilib.SerializedInput`,
                a generic type that can be a :data:`~bodhilib.TextLike`, a list of :data:`~bodhilib.TextLike`,
                or a serialized dict of the object.

        Returns:
            List[:data:`~bodhilib.Node`]: list of :data:`~bodhilib.Node` enriched with :data:`~bodhilib.Embedding`
        """

    @property
    @abc.abstractmethod
    def dimension(self) -> int:
        """Dimension of the embeddings."""


class BaseEmbedder(Embedder):
    """BaseEmbedder provides a simpler method for implementing Embedders.

    BaseEmbedder overrides the abstract Embedder method :func:`~bodhilib.Embedder.embed`,
    massages the data and converts the input to a list of :class:`~bodhilib.Node`,
    and passes to the abstract method to implement :func:`~bodhilib.BaseEmbedder._embed`.
    """

    def embed(self, inputs: SerializedInput) -> List[Node]:
        """Embed a :data:`~bodhilib.SerializedInput` using the embedder service.

        Massages the data and converts the input to a list of :class:`~bodhilib.Node`. Then passes to
        the :func:`~bodhilib.BaseEmbedder._embed` for processing.

        Args:
            inputs (:data:`~bodhilib.SerializedInput`): takes input as :data:`~bodhilib.SerializedInput`,
                a generic type that can be a :data:`~bodhilib.TextLike`, a list of :data:`~bodhilib.TextLike`,
                or a serialized dict of the object.

        Returns:
            List[:data:`~bodhilib.Node`]: list of :class:`~bodhilib.Node` enriched with :data:`~bodhilib.Embedding`
        """
        nodes = to_node_list(inputs)
        return self._embed(nodes)

    @abc.abstractmethod
    def _embed(self, nodes: List[Node]) -> List[Node]:
        """Embed a list of :class:`~bodhilib.Node` using the embedder service.

        Args:
            nodes (List[:class:`~bodhilib.Node`]): list of :class:`~bodhilib.Node` to embed

        Returns:
            List[:data:`~bodhilib.Node`]: list of :class:`~bodhilib.Node` enriched with :data:`~bodhilib.Embedding`
        """


# endregion
# region llm
#######################################################################################################################
class LLM(abc.ABC):
    """Abstract Base Class LLM defines the common interface implemented by all LLM implementations."""

    @abc.abstractmethod
    def generate(
        self,
        prompt_input: SerializedInput,
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Union[Prompt, PromptStream]:
        """Base class :func:`~bodhilib.LLM.generate` method interface common to all LLM service implementation.

        Takes in :data:`bodhilib.SerializedInput`, a flexible input supporting from plain string,
        or a :class:`~bodhilib.Prompt` object, or a dict representation of Prompt.
        Returns the response from LLM service as :class:`~bodhilib.Prompt` object with `source="output"`.

        Args:
            prompts (:data:`bodhilib.SerializedInput`): input to the LLM service
            stream (bool): whether to stream the response from the LLM service
            temperature (Optional[float]): temperature or randomness of the generation
            top_p (Optional[float]): token consideration probability top_p for the generation
            top_k (Optional[int]): token consideration number top_k for the generation
            n (Optional[int]): number of responses to generate
            stop (Optional[List[str]]): list of stop tokens to stop the generation
            max_tokens (Optional[int]): maximum number of tokens to generate
            presence_penalty (Optional[float]): presence penalty for the generation, between -2 and 2
            frequency_penalty (Optional[float]): frequency penalty for the generation, between -2 and 2
            user (Optional[str]): user making the request, for monitoring purpose
            kwargs (Dict[str, Any]): pass through arguments for the LLM service

        Returns:
            :class:`~bodhilib.Prompt`: a Prompt object, if stream is False
            Iterator[:class:`~bodhilib.Prompt`]: an iterator of Prompt objects, if stream is True
        """


class BaseLLM(LLM):
    """BaseLLM provides a simpler method for implementing LLM.

    Class implementing :class:`~bodhilib.LLM` should extend this base class.
    In case the abstract method changes in :class:`~bodhilib.LLM`,
    this base class tries to safely adapt new changes to old interface.
    """

    def generate(
        self,
        prompt_input: SerializedInput,
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Union[Prompt, PromptStream]:
        prompts = to_prompt_list(prompt_input)
        return self._generate(
            prompts,
            stream=stream,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            n=n,
            stop=stop,
            max_tokens=max_tokens,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            user=user,
            **kwargs,
        )

    @abc.abstractmethod
    def _generate(
        self,
        prompts: List[Prompt],
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Union[Prompt, PromptStream]:
        """Generate a response from the LLM service given a List of Prompt.

        Args:
            prompts (List[:data:`bodhilib.Prompt`]): list of prompts as input to LLM service
            stream (bool): whether to stream the response from the LLM service
            temperature (Optional[float]): temperature or randomness of the generation
            top_p (Optional[float]): token consideration probability top_p for the generation
            top_k (Optional[int]): token consideration number top_k for the generation
            n (Optional[int]): number of responses to generate
            stop (Optional[List[str]]): list of stop tokens to stop the generation
            max_tokens (Optional[int]): maximum number of tokens to generate
            presence_penalty (Optional[float]): presence penalty for the generation, between -2 and 2
            frequency_penalty (Optional[float]): frequency penalty for the generation, between -2 and 2
            user (Optional[str]): user making the request, for monitoring purpose
            kwargs (Dict[str, Any]): pass through arguments for the LLM service

        Returns:
            :class:`~bodhilib.Prompt`: a Prompt object, if stream is False
            Iterator[:class:`~bodhilib.Prompt`]: an iterator of Prompt objects, if stream is True
        """
        raise NotImplementedError(f"_generate method is not implemented by class {self.__class__.__name__}")


# endregion
# region vector db
#######################################################################################################################
class VectorDBError(Exception):
    """Error to wrap any exception raised by the vector database client."""


class VectorDB(abc.ABC):
    """VectorDB defines the interface for a vector database client."""

    @abc.abstractmethod
    def ping(self) -> bool:
        """Check the connection to the vector db.

        Returns:
            True if the database is up and running.

        Raises:
            VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """

    @abc.abstractmethod
    def connect(self) -> bool:
        """Establishes a connection with the vector db.

        Returns:
            True if the connection with vector db is successful.

        Raises:
            VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """

    @abc.abstractmethod
    def close(self) -> bool:
        """Closes the connection with the vector db.

        Returns:
            True if the connection with vector db is closed successfully.

        Raises:
            VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """

    @abc.abstractmethod
    def get_collections(self) -> List[str]:
        """List all the vector databases.

        Returns:
            List of vector database names.

        Raises:
            VectorDBError: Wraps any database list error raised by the underlying database.
        """

    @abc.abstractmethod
    def create_collection(
        self,
        collection_name: str,
        dimension: int,
        distance: Union[str, Distance],
        **kwargs: Dict[str, Any],
    ) -> bool:
        """Create a collection in vector database.

        Returns:
            True if the database is created successfully.

        Raises:
            VectorDBError: Wraps any database create error raised by the underlying database.
        """

    @abc.abstractmethod
    def delete_collection(self, collection_name: str, **kwargs: Dict[str, Any]) -> bool:
        """Deletes a collection in vector database.

        Returns:
            True if the database is deleted successfully.

        Raises:
            VectorDBError: Wraps any database delete error raised by the underlying client.
        """

    @abc.abstractmethod
    def upsert(self, collection_name: str, nodes: List[Node]) -> List[Node]:
        """Insert or update a node into the database along with metadata.

        This updates the nodes with record ids and other DB information, similar to using an ORM.

        Returns:
            List of record ids for the inserted embeddings.

        Raises:
            VectorDBError: Wraps any database delete error raised by the underlying client.
        """

    @abc.abstractmethod
    def query(
        self, collection_name: str, embedding: Embedding, filter: Optional[Dict[str, Any]], **kwargs: Dict[str, Any]
    ) -> List[Node]:
        """Search for the nearest vectors in the database.

        Returns:
            List of nodes with metadata.

        Raises:
            VectorDBError: Wraps any database delete error raised by the underlying client.
        """


# endregion
# region plugin
#######################################################################################################################
# PromptSource
PS = TypeVar("PS", bound=PromptSource)
"""TypeVar for PromptSource."""


def get_prompt_source(
    service_name: str,
    *,
    oftype: Optional[Type[PS]] = None,
    publisher: Optional[str] = "bodhiext",
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> PS:
    """Get an instance of PromptSource for given arguments."""
    if oftype is None:
        return_type: Type[Any] = PromptSource
    else:
        return_type = oftype

    manager = PluginManager.instance()
    prompt_source: PS = manager.get(
        service_name=service_name,
        service_type="prompt_source",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(PS, prompt_source)


# DataLoader
DL = TypeVar("DL", bound=DataLoader)
"""TypeVar for DataLoader."""


def get_data_loader(
    service_name: str,
    *,
    oftype: Optional[Type[DL]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> DL:
    """Get an instance of data loader for given arguments.

    Given the service name, publisher (optional) and version(optional),
    return the registered data loader oftype (optional).

    Args:
        service_name (str): name of the service, e.g. "file", "notion", "s3"
        oftype (Optional[Type[T]]): if the type of data loader is known, pass the type in argument `oftype`,
            the data loader is cast to `oftype` and returned for better IDE support.
        publisher (Optional[str]): publisher or developer of the data loader plugin, e.g. "bodhilib","<github-username>"
        version (Optional[str]): version of the data loader
        **kwargs (Dict[str, Any]): pass through arguments for the data loader, e.g. aws_access_key_id, notion_db etc.

    Returns:
        DL (:data:`~bodhilib.DL` | :class:`~bodhilib.DataLoader`): an instance of DataLoader service
            of type `oftype`, if oftype is passed, else of type :class:`~bodhilib.DataLoader`

    Raises:
        TypeError: if the type of data loader is not oftype
    """
    if oftype is None:
        return_type: Type[Any] = DataLoader
    else:
        return_type = oftype

    manager = PluginManager.instance()
    data_loader: DL = manager.get(
        service_name=service_name,
        service_type="data_loader",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(DL, data_loader)


def list_data_loaders() -> List[Service]:
    """List all data loaders installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("data_loader")


# Splitter
S = TypeVar("S", bound=Splitter)
"""TypeVar for Splitter."""


def get_splitter(
    service_name: str,
    *,
    oftype: Optional[Type[S]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> S:
    """Get an instance of splitter for given arguments.

    Returns:
        S (:data:`~bodhilib.S` | :class:`~bodhilib.Splitter`): an instance of Splitter service
    """
    if oftype is None:
        return_type: Type[Any] = Splitter
    else:
        return_type = oftype

    manager = PluginManager.instance()
    splitter: S = manager.get(
        service_name=service_name,
        service_type="splitter",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(S, splitter)


# Embedder
E = TypeVar("E", bound=Embedder)
"""TypeVar for Embedder."""


def get_embedder(
    service_name: str,
    *,
    oftype: Optional[Type[E]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> E:
    """Get an instance of embedder given the service name, publisher (optional) and version(optional).

    Args:
        service_name (str): name of the service, e.g. "sentence-transformers" etc.
        oftype (Optional[Type[T]]): if the type of embedder is known, pass the type in argument `oftype`,
            the embedder is cast to `oftype` and returned for better IDE support.
        publisher (Optional[str]): publisher or developer of the embedder plugin, e.g. "bodhilib","<github-username>"
        version (Optional[str]): version of the embedder
        **kwargs (Dict[str, Any]): pass through arguments for the embedder, e.g. dimension etc.

    Returns:
        E (:data:`~bodhilib.E` | :class:`~bodhilib.Embedder`): an instance of Embedder service
            of type `oftype`, if oftype is passed, else of type :class:`~bodhilib.Embedder`

    Raises:
        TypeError: if the type of embedder is not oftype
    """
    if oftype is None:
        return_type: Type[Any] = Embedder
    else:
        return_type = oftype

    manager = PluginManager.instance()
    embedder: E = manager.get(
        service_name, "embedder", oftype=return_type, publisher=publisher, version=version, **kwargs
    )
    return cast(E, embedder)


def list_embedders() -> List[Service]:
    """List all embedders installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("embedder")


# LLM
L = TypeVar("L", bound=LLM)
"""TypeVar for LLM."""


def get_llm(
    service_name: str,
    model: str,
    api_key: Optional[str] = None,
    *,
    oftype: Optional[Type[L]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> L:
    """Get an instance of LLM for the given service name and model.

    Args:
        service_name (str): name of the service, e.g. "openai", "cohere", "anthropic"
        model (str): name of the model, e.g. "chat-gpt-3.5", "command", "claude-2"
        api_key (Optional[str]): API key for the service, if the api_key is not provided,
            it will be looked in the environment variables
        oftype (Optional[Type[T]]): if the type of LLM is known, pass the type in argument `oftype`,
            the LLM is cast to `oftype` and returned for better IDE support.
        publisher (Optional[str]): publisher or developer of the service plugin, e.g. "bodhilib", "<github-username>"
        version (Optional[str]): version of the service
        **kwargs (Dict[str, Any]): pass through arguments for the LLM service, e.g. "temperature", "max_tokens", etc.

    Returns:
        T (:data:`~bodhilib.L` | :class:`~bodhilib.LLM`):
            an instance of LLM service of type `oftype`, if oftype is passed,
            else of type :class:`~bodhilib.LLM`

    Raises:
        TypeError: if the type of LLM is not oftype
    """
    if oftype is None:
        return_type: Type[Any] = LLM
    else:
        return_type = oftype

    manager = PluginManager.instance()
    llm: L = manager.get(
        service_name=service_name,
        service_type="llm",
        oftype=return_type,
        publisher=publisher,
        version=version,
        model=model,
        api_key=api_key,
        **kwargs,
    )
    return cast(L, llm)


def list_llms() -> List[Service]:
    """List all LLM services installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("llm")


# VectorDB
V = TypeVar("V", bound=VectorDB)
"""TypeVar for VectorDB."""


def get_vector_db(
    service_name: str,
    *,
    oftype: Optional[Type[V]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> V:
    """Get an instance of VectorDB for the given service name.

    Returns:
        V (:data:`~bodhilib.V` | :class:`~bodhilib.VectorDB`): an instance of VectorDB service
    """
    if oftype is None:
        return_type: Type[Any] = VectorDB
    else:
        return_type = oftype

    manager = PluginManager.instance()
    vector_db: V = manager.get(
        service_name=service_name,
        service_type="vector_db",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(V, vector_db)


def list_vector_dbs() -> List[Service]:
    """List all VectorDB services installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("vector_db")


# endregion



""":mod:`bodhiext.sentence_transformers` module defines classes and methods for embedder using sentence-transformer."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from bodhilib import BaseEmbedder, Embedding, Node, Service, service_provider

# TODO: Once the module is separate, import this from the extension version file
from bodhilib import __version__ as bodhilib_version
from bodhilib.logging import logger
from sentence_transformers import SentenceTransformer

__version__ = bodhilib_version


class SentenceTransformerEmbedder(BaseEmbedder):
    """Embedder using sentence-transformer library."""

    def __init__(
        self, client: Optional[SentenceTransformer] = None, model: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> None:
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        self.kwargs = kwargs
        self.client: Optional[SentenceTransformer] = None
        if client:
            self.client = client
            return
        if model is None:
            logger.info("No model passed to SentenceTransformer. Using default model 'all-MiniLM-L6-v2'")
            self.model = "all-MiniLM-L6-v2"
        else:
            self.model = model

    def _embed(self, nodes: List[Node]) -> List[Node]:
        """Embeds the nodes using sentence-transformer.

        Args:
            nodes (List[:class:`~bodhilib.Node`]): list of nodes to embed

        Returns:
            List[:class:`~bodhilib.Embedding`]: list of embeddings
        """
        if self.client is None:
            self.client = SentenceTransformer(self.model)
        embeddings: List[Embedding] = self.client.encode([node.text for node in nodes]).tolist()
        for node, embedding in zip(nodes, embeddings):
            node.embedding = embedding
        return nodes

    @property
    def dimension(self) -> int:
        """Dimension of the embeddings.

        Returns:
            int: dimension of the embeddings
        """
        if self.client is None:
            self.client = SentenceTransformer(self.model)
        dimension = self.client.get_sentence_embedding_dimension()
        if dimension is None:
            raise ValueError("Dimension of the model is None.")
        if isinstance(dimension, int):
            return dimension
        raise ValueError(f"Unknown type for dimension, type={type(dimension)}")


def sentence_transformer_builder(
    *,
    service_name: Optional[str] = None,
    service_type: Optional[str] = "embedder",
    client: Optional[SentenceTransformer] = None,
    model: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> SentenceTransformerEmbedder:
    """Returns an instance of sentence transformer builder.

    Args:
        service_name: service name to wrap, should be "sentence_transformers"
        service_type: service of the implementation, should be "embedder"
        client: the client to use for embedding, if not supplied, a new client is created
        model: the LLM model to use for embedding, if not supplied, a default is used
        **kwargs: pass through arguments for the embedder, e.g. dimension etc.
    """
    if service_name != "sentence_transformers":
        raise ValueError(f"Unknown service: {service_name=}")
    if service_type != "embedder":
        raise ValueError(f"Service type not supported: {service_type=}, supported service type: 'embedder'")
    return SentenceTransformerEmbedder(client=client, model=model, **kwargs)


@service_provider
def bodhilib_list_services() -> List[Service]:
    """Return a list of services supported by the plugin."""
    return [
        Service(
            service_name="sentence_transformers",
            service_type="embedder",
            publisher="bodhiext",
            service_builder=sentence_transformer_builder,
            version=__version__,
        )
    ]



"""LLM implementation for Cohere."""
from __future__ import annotations

import os
from typing import Any, Dict, List, Optional

from bodhilib import LLM, Service, service_provider

import cohere

from ._cohere_llm import Cohere
from ._version import __version__


@service_provider
def bodhilib_list_services() -> List[Service]:
    """This function is used by bodhilib to find all services in this module."""
    return [
        Service(
            service_name="cohere",
            service_type="llm",
            publisher="bodhiext",
            service_builder=cohere_llm_service_builder,
            version=__version__,
        )
    ]


def cohere_llm_service_builder(
    *,
    service_name: Optional[str] = None,
    service_type: Optional[str] = "llm",
    client: Optional[cohere.Client] = None,
    model: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> LLM:
    """Returns an instance of Cohere LLM service implementing :class:`bodhilib.LLM`.

    Args:
        service_name: service name to wrap, should be "cohere"
        service_type: service type of the implementation, should be "llm"
        client: Cohere client instance, if not set, it will be created using other parameters
        model: Cohere model identifier
        api_key: api key for Cohere service, if not set, it will be read from environment variable COHERE_API_KEY
    Returns:
        LLM: a service instance implementing :class:`bodhilib.LLM` for the given service and model
    Raises:
        ValueError: if service_name is not "cohere"
        ValueError: if service_type is not "llm"
        ValueError: if model is not set
        ValueError: if api_key is not set, and environment variable COHERE_API_KEY is not set
    """
    # TODO use pydantic for parameter validation
    if service_name != "cohere" or service_type != "llm":
        raise ValueError(
            f"Unknown params: {service_name=}, {service_type=}, supported params: service_name='cohere',"
            " service_type='llm'"
        )
    if api_key is None:
        if os.environ.get("COHERE_API_KEY") is None:
            raise ValueError("environment variable COHERE_API_KEY is not set")
        else:
            api_key = os.environ["COHERE_API_KEY"]
    return Cohere(client=client, model=model, api_key=api_key, **kwargs)



"""mod:`bodhiext.splitter` bodhiext package for splitters."""
import inspect

from ._plugin import bodhilib_list_services as bodhilib_list_services
from ._plugin import text_splitter_service_builder as text_splitter_service_builder
from ._text_splitter import TextSplitter as TextSplitter
from ._version import __version__ as __version__

__all__ = [name for name, obj in globals().items() if not (name.startswith("_") or inspect.ismodule(obj))]

del inspect



from __future__ import annotations

import io
import itertools
import re
import reprlib
import textwrap
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    no_type_check,
)

from jinja2 import Template
from pydantic import BaseModel, Field, validator
from typing_extensions import TypeAlias

# region type aliases
#######################################################################################################################
# Update the documentation in bodhilib.rst file directly. TypeAlias inline documentation not picked up by sphinx.
# Duplicating it here to make it easier when browsing the code.
PathLike: TypeAlias = Union[str, Path]
"""Type alias for Union of :class:`str` and :class:`~pathlib.Path`"""
TextLike: TypeAlias = Union[str, "SupportsText"]
"""Type alias for Union of :class:`str` and protocol :data:`~bodhilib.SupportsText`"""
TextLikeOrTextLikeList: TypeAlias = Union[TextLike, Iterable[TextLike]]
"""Type alias for Union of :data:`~bodhilib.TextLike` or list of :data:`~bodhilib.TextLike`"""
SerializedInput: TypeAlias = Union[TextLikeOrTextLikeList, Dict[str, Any], Iterable[Dict[str, Any]]]
"""Type alias for various inputs that can be passed to the components."""
Embedding: TypeAlias = List[float]
"""Type alias for list of :class:`float`, to indicate the embedding generated
from :class:`~bodhilib.Embedder` operation"""


class SupportsText(Protocol):
    """TextLike is a protocol for types that can be converted to text.

    To support the protocol, the type must have a property `text`.

    Known sub-classes: :class:`~bodhilib.Prompt`, :class:`~bodhilib.Document`, :class:`~bodhilib.Node`
    """

    @property
    def text(self) -> str:
        """Return the content of the object as string."""


def supportstext(obj: object) -> bool:
    """Returns True if the object supports :data:`~bodhilib.SupportsText` protocol."""
    return hasattr(obj, "text")


def istextlike(obj: object) -> bool:
    """Returns True if the object is a :data:`~TextLike`."""
    return isinstance(obj, str) or supportstext(obj)


# endregion
# region utility
#######################################################################################################################
class _StrEnumMixin:
    """Mixin class for string enums, provides __str__ and __eq__ methods."""

    @no_type_check
    def __str__(self) -> str:
        """Returns the string value of the string enum."""
        return self.value

    @no_type_check
    def __eq__(self, other: Any) -> bool:
        """Compares this string enum to other string enum or string values."""
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, type(self)):
            return self.value == other.value
        return False

    @no_type_check
    @classmethod
    def membersstr(cls) -> List[str]:
        return [e.value for e in cls.__members__.values()]


_EnumT = TypeVar("_EnumT", bound=Enum)
"""TypeVar for Enum type."""


def _strenum_validator(enum_cls: Type[_EnumT], value: Any) -> _EnumT:
    """Converts a string value to an enum value."""
    if isinstance(value, str):
        try:
            return enum_cls[value.upper()]
        except KeyError as e:
            allowed_values = [e.value for e in enum_cls]
            raise ValueError(f"Invalid value for {enum_cls.__name__}. Allowed values are {allowed_values}.") from e
    elif isinstance(value, enum_cls):
        return value
    else:
        raise ValueError(f"Invalid type for value, {type(value)=}")


# endregion
# region value objects
#######################################################################################################################
class Role(_StrEnumMixin, str, Enum):
    """Role of the prompt.

    Used for fine-grain control over "role" instructions to the LLM service.
    Can be one of - *'system', 'ai', or 'user'*.
    """

    SYSTEM = "system"
    AI = "ai"
    USER = "user"


class Source(_StrEnumMixin, str, Enum):
    """Source of the prompt.

    If the prompt is given as input by the user, then *source="input"*,
    or if the prompt is generated as response by the LLM service, then *source="output"*.
    """

    INPUT = "input"
    OUTPUT = "output"


class Distance(_StrEnumMixin, str, Enum):
    """Vector Distance Method."""

    COSINE = "cosine"
    EUCLIDEAN = "euclidean"
    DOT_PRODUCT = "dot_product"


# endregion
# region prompt
#######################################################################################################################
class Prompt(BaseModel):
    """Prompt encapsulating input/output schema to interact with LLM service."""

    text: str
    """The text or content or input component of the prompt."""

    role: Role = Role.USER
    """The role of the prompt.

    Defaults to :obj:`Role.USER`."""

    source: Source = Source.INPUT
    """The source of the prompt.

    Defaults to :obj:`Source.INPUT`."""

    # overriding __init__ to provide positional argument construction for prompt. E.g. `Prompt("text")`
    def __init__(
        self,
        text: str,
        role: Optional[Union[Role, str]] = Role.USER,
        source: Optional[Union[Source, str]] = Source.INPUT,
    ):
        """Initialize a prompt.

        Args:
            text (str): text of the prompt
            role (Role): role of the prompt.

                Role can be given as one of the allowed string value ["system", "ai", "user"]
                or as a Role enum [:obj:`Role.SYSTEM`, :obj:`Role.AI`, :obj:`Role.USER`].

                The string is converted to Role enum. If the string value is not one of the allowed values,
                then a ValueError is raised.
            source (Source): source of the prompt.

                Source can be given as a one of the allowed string value ["input", "output"]
                or as a Source enum [:obj:`Source.INPUT`, :obj:`Source.OUTPUT`].

                The string is converted to Source enum. If the string value is not one of the allowed values,
                then a ValueError is raised.

        Raises:
            ValueError: If the role or source is not one of the allowed values.
        """
        role = role or Role.USER
        source = source or Source.INPUT
        super().__init__(text=text, role=role, source=source)

    @validator("role", pre=True, always=True)
    def validate_role(cls, value: Any) -> Role:
        return _strenum_validator(Role, value)

    @validator("source", pre=True, always=True)
    def validate_source(cls, value: Any) -> Source:
        return _strenum_validator(Source, value)

    def isstream(self) -> bool:
        """To check if this is a prompt stream.

        Returns:
            bool: False as this is not a prompt stream.
        """
        return False


T = TypeVar("T")
"""TypeVar for LLM Response Type."""


class PromptStream(Iterator[Prompt]):
    """Iterator over a stream of prompts.

    Used by LLMs to wrap the stream response to an iterable over prompts.
    """

    def __init__(self, api_response: Iterable[T], transformer: Callable[[T], Prompt]):
        """Initialize a prompt stream.

        Args:
            api_response (Iterable[T]): LLM API Response of generic type :data:`~bodhilib.T` as Iterable
            transformer (Callable[[T], Prompt]): Transformer function to convert API response to Prompt
        """
        self.api_response = iter(api_response)
        self.transformer = transformer
        self.output = io.StringIO()
        self.role: Optional[str] = None

    def __iter__(self) -> Iterator[Prompt]:
        """Returns the iterator object itself."""
        return self

    def __next__(self) -> Prompt:
        """Returns the next item from the iterator as Prompt object."""
        try:
            chunk_response = next(self.api_response)
        except StopIteration as e:
            raise StopIteration from e
        prompt = self.transformer(chunk_response)
        if self.role is None:
            self.role = prompt.role
        self.output.write(prompt.text)
        return prompt

    def isstream(self) -> bool:
        """To check if this is a prompt stream.

        Returns:
            bool: False as this is not a prompt stream.
        """
        return True

    @property
    def text(self) -> str:
        # TODO change it to first read all the stream content, and then return the text
        # Create another property to return so-far accumulated text
        """Returns the text accumulated over the stream of responses."""
        return self.output.getvalue()


def prompt_user(text: str) -> Prompt:
    """Factory method to generate user prompt from string.

    Args:
        text: text of the prompt

    Returns:
        Prompt: Prompt object generated from the text. Defaults role="user" and source="input".
    """
    return Prompt(text=text, role=Role.USER, source=Source.INPUT)


def prompt_system(text: str) -> Prompt:
    """Factory method to generate system prompt from string.

    Args:
        text: text of the prompt

    Returns:
        Prompt: Prompt object generated from the text. Defaults role="system" and source="input".
    """
    return Prompt(text=text, role=Role.SYSTEM, source=Source.INPUT)


def prompt_output(text: str) -> Prompt:
    """Factory method to generate output prompts.

    Generates a prompt with source="output". Mainly by LLMs to generate output prompts.
    """
    return Prompt(text=text, role=Role.AI, source=Source.OUTPUT)


# endregion
# region prompt template
#######################################################################################################################

TemplateFormat = Literal["fstring", "jinja2", "bodhilib-fstring", "bodhilib-jinja2"]


class PromptTemplate(BaseModel):
    """PromptTemplate used for generating prompts using a template."""

    id: Optional[str]
    """Optional identifier for prompt template"""

    template: str
    """Template for generating prompts."""

    role: Role = Role.USER
    """Role of the prompt."""

    source: Source = Source.INPUT
    """Source of the prompt."""

    format: TemplateFormat = "fstring"
    """Template format to use for rendering."""

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata associated with the template."""

    vars: Dict[str, Any] = Field(default_factory=dict)
    """The variables to be used for rendering the template."""

    # overriding __init__ to provide positional argument construction for prompt template.
    # E.g. `PromptTemplate("my template {context}")`
    def __init__(
        self,
        template: str,
        *,
        id: Optional[str] = None,
        role: Optional[Role] = None,
        source: Optional[Source] = None,
        format: Optional[TemplateFormat] = "fstring",
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        """Initializes a prompt template.

        Args:
            template: template string
            id: optional identifier for the template
            role: role of the prompt, defaults to "user"
            source: source of the prompt, defaults to "input"
            format: format to use for rendering the template.
            metadata: the metadata associated with the prompt template
            **kwargs: additional arguments to be used for rendering the template
        """
        role = role or Role.USER
        source = source or Source.INPUT
        super().__init__(
            template=template, id=id, role=role, source=source, format=format, metadata=metadata or {}, vars=kwargs
        )

    def to_prompts(self, **kwargs: Dict[str, Any]) -> List[Prompt]:
        """Converts the PromptTemplate into a Prompt.

        Args:
            kwargs: all variables to be used for rendering the template

        Returns:
            Prompt: prompt generated from the template
        """
        all_args = {**self.vars, **kwargs}
        all_args = {k: v for k, v in all_args.items() if v is not None}
        if self.format == "fstring":
            return [Prompt(self.template.format(**all_args), role=self.role, source=self.source)]
        if self.format == "jinja2":
            template = Template(textwrap.dedent(self.template))
            result = template.render(**all_args)
            return [Prompt(result, role=self.role, source=self.source)]
        if self.format.startswith("bodhilib-"):
            return self._bodhilib_template_to_prompt(**all_args)
        raise ValueError(
            f"Unknown format {self.format}, "
            "allowed values: ['fstring', 'jinja2', 'bodhilib-fstring', 'bodhilib-jinja2']"
        )

    def _bodhilib_template_to_prompt(self, **kwargs: Dict[str, Any]) -> List[Prompt]:
        prompt_fields = ["text", "role", "source"]
        prompt_fields_matcher = "^" + "|".join(prompt_fields) + ":"
        lines = self.template.splitlines(keepends=True)
        result: List[Prompt] = []
        prompt: Dict[str, Any] = {"text": []}
        text_start = False
        for line in lines:
            if re.match(prompt_fields_matcher, line):
                field, value = line.split(":")
                if field == "text":
                    text_start = True
                    prompt["text"].append(value)
                else:
                    text_start = False
                    prompt[field] = value.strip()
                continue
            if line.startswith("---"):
                if not prompt["text"]:
                    text_start = False
                    prompt = {"text": []}
                    continue
                p = self._build_prompt(prompt, **kwargs)
                result.append(p)
                text_start = False
                prompt = {"text": []}
                continue
            if text_start:
                prompt["text"].append(line)
        if prompt["text"]:
            p = self._build_prompt(prompt, **kwargs)
            result.append(p)
        return result

    def _build_prompt(self, prompt: Dict[str, Any], **kwargs: Dict[str, Any]) -> Prompt:
        template = "".join(prompt.pop("text"))
        if self.format == "bodhilib-fstring":
            text = template.format(**kwargs)
        elif self.format == "bodhilib-jinja2":
            jinja_template = Template(template, keep_trailing_newline=True)
            text = jinja_template.render(**kwargs)
        else:
            raise ValueError("Unknown format {self.format}, allowed values: ['bodhilib-fstring', 'bodhilib-jinja2']")
        return Prompt(text, **prompt)


# TODO deprecate and remove
def prompt_with_examples(template: str, **kwargs: Dict[str, Any]) -> PromptTemplate:
    """Factory method to generate a prompt template with examples.

    Prompt uses `jinja2` template engine to generate prompt with examples.

    Args:
        template: a `jinja2` compliant template string to loop through examples
        **kwargs: additional arguments to be used for rendering the template.
            Can also contain `role` and `source` to override the default values.

    Returns:
        PromptTemplate: configured prompt template to generate prompt with examples
    """
    # pop role from kwargs or get None
    role = kwargs.pop("role", None)
    source = kwargs.pop("source", None)
    return PromptTemplate(template, role=role, source=source, format="jinja2", **kwargs)  # type: ignore


# TODO deprecate and remove
def prompt_with_extractive_qna(template: str, contexts: List[TextLike], **kwargs: Dict[str, Any]) -> PromptTemplate:
    """Factory method to generate a prompt template for extractive QnA.

    Args:
        template: a `jinja2` compliant template string to loop through examples
        **kwargs: additional arguments to be used for rendering the template.
            Can also contain `role` and `source` to override the default values.

    Returns:
        PromptTemplate: configured prompt template to generate prompt with examples
    """
    # pop role from kwargs or get None
    role = kwargs.pop("role", None)
    source = kwargs.pop("source", None)
    return PromptTemplate(
        template, role=role, source=source, format="jinja2", contexts=contexts, **kwargs  # type: ignore
    )


# endregion
# region document
#######################################################################################################################
class Document(BaseModel):
    """Document defines the basic interface for a processible resource.

    Primarily contains text (content) and metadata.
    """

    text: str
    """Text content of the document."""

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata associated with the document. e.g. filename, dirname, url etc."""

    def __repr__(self) -> str:
        """Returns a string representation of the document."""
        return f"Document(text={reprlib.repr(self.text)}, metadata={reprlib.repr(self.metadata)})"


class Node(BaseModel):
    """Node defines the basic data structure for a processible resource.

    It contains a unique identifier, content text, metadata associated with its sources,
    and embeddings.
    """

    id: Optional[str] = None
    """Unique identifier for the node.

    Generated during the document split operation, or retrieved from doc/vector database at the time of query."""

    text: str
    """Text content of the document."""

    parent: Optional[Document] = None
    """Metadata associated with the document. e.g. filename, dirname, url etc."""

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata associated with the node. This is also copied over from parent when splitting Document."""

    embedding: Optional[Embedding] = None

    def __repr__(self) -> str:
        """Returns a string representation of the document."""
        return f"Node(id={self.id}, text={reprlib.repr(self.text)}, parent={repr(self.parent)})"


# endregion
# region model converters
#######################################################################################################################
def to_document(textlike: TextLike) -> Document:
    """Converts a :data:`~bodhilib.TextLike` to :class:`~bodhilib.Document`."""
    if isinstance(textlike, Document):
        return textlike
    elif isinstance(textlike, str):
        return Document(text=textlike)
    elif supportstext(textlike):
        return Document(text=textlike.text)
    raise ValueError(f"Cannot convert type {type(textlike)} to Document.")


def to_prompt(textlike: TextLike) -> Prompt:
    """Converts a :data:`~TextLike` to :class:`~Prompt`."""
    if isinstance(textlike, Prompt):
        return textlike
    elif isinstance(textlike, str):
        return Prompt(text=textlike)
    elif supportstext(textlike):
        return Prompt(text=textlike.text)
    raise ValueError(f"Cannot convert type {type(textlike)} to Prompt.")


def to_node(textlike: TextLike) -> Node:
    """Converts a :data:`~TextLike` to :class:`~Node`."""
    if isinstance(textlike, Node):
        return textlike
    elif isinstance(textlike, str):
        return Node(text=textlike)
    elif supportstext(textlike):
        return Node(text=textlike.text)
    raise ValueError(f"Cannot convert type {type(textlike)} to Node.")


def to_text(textlike: TextLike) -> str:
    """Converts a :data:`~TextLike` to string."""
    if isinstance(textlike, str):
        return textlike
    if supportstext(textlike):
        return textlike.text
    raise ValueError(f"Cannot convert type {type(textlike)} to text.")


def to_prompt_list(inputs: SerializedInput) -> List[Prompt]:
    """Converts a :data:`~bodhilib.SerializedInput` to list of :class:`~Prompt`."""
    if istextlike(inputs):
        return [to_prompt(cast(TextLike, inputs))]  # cast to fix mypy warning
    elif isinstance(inputs, dict):
        return [Prompt(**inputs)]
    elif isinstance(inputs, Iterable):
        result = [to_prompt_list(textlike) for textlike in inputs]
        return list(itertools.chain(*result))
    else:
        return [to_prompt(inputs)]


def to_document_list(inputs: SerializedInput) -> List[Document]:
    """Converts a :data:`~bodhilib.SerializedInput` to list of :class:`~Document`."""
    if istextlike(inputs):
        return [to_document(cast(TextLike, inputs))]  # cast to fix mypy warning
    elif isinstance(inputs, dict):
        return [Document(**inputs)]
    if isinstance(inputs, Iterable):
        result = [to_document_list(input) for input in inputs]
        return list(itertools.chain(*result))
    else:
        return [to_document(inputs)]


def to_node_list(inputs: SerializedInput) -> List[Node]:
    """Converts a :data:`~bodhilib.SerializedInput` to list of :class:`~Node`."""
    if (
        not isinstance(inputs, BaseModel)  # BaseModel is Iterable
        and isinstance(inputs, Iterable)  # if is list
        and all(isinstance(input, Node) for input in inputs)  # and if all are Node instance
    ):
        return inputs  # type: ignore
    if istextlike(inputs):
        return [to_node(cast(TextLike, inputs))]  # cast to fix mypy warning
    elif isinstance(inputs, dict):
        return [Node(**inputs)]
    elif isinstance(inputs, Iterable):
        result = [to_node_list(input) for input in inputs]
        return list(itertools.chain(*result))
    else:
        return [to_node(inputs)]


# endregion
# region utils
#######################################################################################################################


# endregion



"""OpenAI LLM module."""
from __future__ import annotations

from typing import Any, Dict, Iterable, List, NoReturn, Optional, Union

from bodhilib import LLM, Prompt, PromptStream, Role, SerializedInput, prompt_output, to_prompt_list

import openai
from openai.openai_response import OpenAIResponse


class OpenAIChat(LLM):
    """OpenAI Chat API implementation for :class:`bodhilib.LLM`."""

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        self.kwargs = kwargs

    def generate(
        self,
        prompt_input: SerializedInput,
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Union[Prompt, PromptStream]:
        prompts = to_prompt_list(prompt_input)
        all_args = {
            **self.kwargs,
            "stream": stream,
            "temperature": temperature,
            "top_p": top_p,
            "n": n,
            "stop": stop,
            "max_tokens": max_tokens,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
            "logit_bias": kwargs.get("logit_bias", None),
            "user": user,
            **kwargs,
        }
        all_args = {k: v for k, v in all_args.items() if v is not None}
        if "model" not in all_args:
            raise ValueError("parameter model is required")
        messages = self._to_messages(prompts)
        response = openai.ChatCompletion.create(messages=messages, **all_args)
        if "stream" in all_args and all_args["stream"]:
            return PromptStream(response, _chat_response_to_prompt_transformer)
        content = response["choices"][0]["message"]["content"]
        return prompt_output(content)

    def _to_messages(self, prompts: List[Prompt]) -> List[Dict[str, str]]:
        role_lookup = {Role.SYSTEM.value: "system", Role.AI.value: "assistant", Role.USER.value: "user"}
        return [{"role": role_lookup[p.role.value], "content": p.text} for p in prompts]

    def __call__(self, *args: Iterable[Any], **kwargs: Dict[str, Any]) -> NoReturn:
        raise TypeError(f"'{type(self).__name__}' object is not callable, did you mean to call 'generate'?")


class OpenAIText(LLM):
    """OpenAI Text API implementation for :class:`bodhilib.LLM`."""

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        self.kwargs = kwargs

    def generate(
        self,
        prompt_input: SerializedInput,
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Any,
    ) -> Union[Prompt, PromptStream]:
        prompts = to_prompt_list(prompt_input)
        prompt = self._to_prompt(prompts)
        all_args = {
            **self.kwargs,
            "stream": stream,
            "suffix": kwargs.pop("suffix", None),
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "n": n,
            "logprobs": kwargs.pop("logprobs", None),
            "echo": kwargs.pop("echo", None),
            "stop": stop,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
            "best_of": kwargs.pop("best_of", None),
            "logit_bias": kwargs.get("logit_bias", None),
            "user": user,
            **kwargs,
        }
        all_args = {k: v for k, v in all_args.items() if v is not None}
        if "model" not in all_args:
            raise ValueError("parameter model is required")
        response = openai.Completion.create(prompt=prompt, **all_args)
        if "stream" in all_args and all_args["stream"]:
            return PromptStream(response, _text_response_to_prompt_transfromer)
        return _text_response_to_prompt_transfromer(response)

    def _to_prompt(self, prompts: List[Prompt]) -> str:
        return "\n".join([p.text for p in prompts])

    def __call__(self, *args: Iterable[Any], **kwargs: Dict[str, Any]) -> NoReturn:
        raise TypeError(f"'{type(self).__name__}' object is not callable, did you mean to call 'generate'?")


def _chat_response_to_prompt_transformer(response: OpenAIResponse) -> Prompt:
    result = response["choices"][0]
    content = "" if result["finish_reason"] else result["delta"]["content"]
    return prompt_output(content)


def _text_response_to_prompt_transfromer(response: OpenAIResponse) -> Prompt:
    result = response["choices"][0]["text"]
    return prompt_output(result)



import re
from typing import Any, Dict, List

from ._models import PromptTemplate

BREAK_TEMPLATE = "+++"
BREAK_PROMPT = "---"
metadata_fields = ["tags", "format"]
metadata_field_pattern = "^" + "|".join(metadata_fields) + ":"


def parse_prompt_template(text: str) -> List[PromptTemplate]:
    """Parses a bodhilib-* template format to PromptTemplate."""
    lines = text.splitlines(keepends=True)
    result: List[PromptTemplate] = []
    prompt_template: Dict[str, Any] = {"template": [], "metadata": {}}
    metadata_section = False
    for line in lines:
        if re.match(metadata_field_pattern, line):
            field_name, field_vals = line.split(":")
            field_vals = field_vals.strip()
            if field_name == "tags":
                field_vals = field_vals.split(",")  # type: ignore
                field_vals = [val.strip() for val in field_vals]  # type: ignore
            prompt_template["metadata"][field_name] = field_vals
            metadata_section = True
            continue
        if line.startswith(BREAK_PROMPT) and metadata_section:
            # gulp the line and continue
            metadata_section = False
            prompt_template["template"] = []  # reset the template content
            continue
        if line.startswith(BREAK_TEMPLATE):
            if "template" in prompt_template and prompt_template["template"]:
                result.append(_build_prompt_template(prompt_template))
            prompt_template = {"template": [], "metadata": {}}
            continue
        prompt_template["template"].append(line)
    if "template" in prompt_template and prompt_template["template"]:
        result.append(_build_prompt_template(prompt_template))
    return result


def _build_prompt_template(prompt_template: Dict[str, Any]) -> PromptTemplate:
    template = prompt_template.pop("template")
    prompt_template["template"] = "".join(template)
    if "format" not in prompt_template["metadata"]:
        prompt_template["format"] = "bodhilib-fstring"
    else:
        prompt_template["format"] = prompt_template["metadata"].pop("format")
    pt = PromptTemplate(**prompt_template)
    return pt



from typing import Any, Dict, List, Optional

from bodhilib import Service, service_provider

from ._text_splitter import TextSplitter
from ._version import __version__


def text_splitter_service_builder(
    *,
    service_name: Optional[str] = "text_splitter",
    service_type: Optional[str] = "splitter",
    publisher: Optional[str] = "bodhiext",
    version: Optional[str] = None,
    max_len: Optional[int] = None,
    min_len: Optional[int] = None,
    overlap: Optional[int] = None,
    eos_patterns: Optional[List[str]] = None,
    eow_patterns: Optional[List[str]] = None,
    **kwargs: Dict[str, Any],
) -> TextSplitter:
    """Service builder for text splitter."""
    if service_name != "text_splitter" or service_type != "splitter" or publisher != "bodhiext":
        raise ValueError(
            f"Unknown service: {service_name=}, {service_type=}, {publisher=}, "
            "supported service: service_name='text_splitter', service_type='splitter', publisher='bodhiext'"
        )
    all_args = {
        "max_len": max_len,
        "min_len": min_len,
        "overlap": overlap,
        "eos_patterns": eos_patterns,
        "eow_patterns": eow_patterns,
        **kwargs,
    }
    all_args = {k: v for k, v in all_args.items() if v is not None}
    # TODO: Add validation for all_args, extra args fails the constructor
    return TextSplitter(**all_args)  # type: ignore


@service_provider
def bodhilib_list_services() -> List[Service]:
    """Return a list of services supported by the bodhiext data_loaders package."""
    return [
        Service(
            service_name="text_splitter",
            service_type="splitter",
            publisher="bodhiext",
            service_builder=text_splitter_service_builder,
            version=__version__,
        )
    ]



from typing import List

from bodhilib import Service, service_provider

from ._file import file_loader_service_builder as file_loader_service_builder
from ._version import __version__


@service_provider
def bodhilib_list_services() -> List[Service]:
    """Return a list of services supported by the bodhiext data_loaders package."""
    return [
        Service(
            service_name="file",
            service_type="data_loader",
            publisher="bodhiext",
            service_builder=file_loader_service_builder,
            version=__version__,
        )
    ]



"""LLM implementation for Cohere."""
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from bodhilib import LLM, Prompt, PromptStream, SerializedInput, to_prompt_list, prompt_output

import cohere
from cohere.responses.generation import StreamingText


class Cohere(LLM):
    """Cohere API implementation for :class:`~bodhilib.LLM`."""

    def __init__(
        self,
        client: Optional[cohere.Client] = None,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ):
        all_args = {"model": model, "api_key": api_key, **kwargs}
        self.kwargs = {k: v for k, v in all_args.items() if v is not None}

        if client:
            self.client = client
        else:
            allowed_args = [
                "api_key",
                "num_workers",
                "request_dict",
                "check_api_key",
                "client_name",
                "max_retries",
                "timeout",
                "api_url",
            ]
            args = {k: v for k, v in self.kwargs.items() if k in allowed_args}
            self.client = cohere.Client(**args)

    def generate(
        self,
        prompt_input: SerializedInput,
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Union[Prompt, PromptStream]:
        prompts = to_prompt_list(prompt_input)
        if len(prompts) == 0:
            raise ValueError("Prompt is empty")
        input = self._to_cohere_prompt(prompts)
        if input == "":
            raise ValueError("Prompt is empty")
        all_args = {
            **self.kwargs,
            "stream": stream,
            "num_generations": n,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "p": top_p,
            "k": top_k,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "stop_sequences": stop,
            "user": user,
            **kwargs,
        }
        all_args = {k: v for k, v in all_args.items() if v is not None}
        if "model" not in all_args:
            raise ValueError("parameter model is required")

        allowed_args = [
            "prompt",
            "prompt_vars",
            "model",
            "preset",
            "num_generations",
            "max_tokens",
            "temperature",
            "k",
            "p",
            "frequency_penalty",
            "presence_penalty",
            "end_sequences",
            "stop_sequences",
            "return_likelihoods",
            "truncate",
            "logit_bias",
            "stream",
            "user",
        ]
        args = {k: v for k, v in all_args.items() if k in allowed_args}

        response = self.client.generate(input, **args)

        if "stream" in all_args and all_args["stream"]:
            return PromptStream(response, _cohere_stream_to_prompt_transformer)
        text = response.generations[0].text
        return prompt_output(text)

    def _to_cohere_prompt(self, prompts: List[Prompt]) -> str:
        return "\n".join([p.text for p in prompts])


def _cohere_stream_to_prompt_transformer(chunk: StreamingText) -> Prompt:
    return prompt_output(chunk.text)



""":mod:`bodhiext.prompt_source` bodhiext package for prompt source."""
import inspect

from ._prompt_source import LocalDirectoryPromptSource as LocalDirectoryPromptSource
from ._prompt_source import bodhi_prompt_source_builder as bodhi_prompt_source_builder
from ._prompt_source import bodhilib_list_services as bodhilib_list_services
from ._version import __version__ as __version__

__all__ = [name for name, obj in globals().items() if not (name.startswith("_") or inspect.ismodule(obj))]

del inspect



"""OpenAI API operations."""
import inspect

from ._openai_llm import OpenAIChat as OpenAIChat
from ._openai_llm import OpenAIText as OpenAIText
from ._openai_plugin import bodhilib_list_services as bodhilib_list_services
from ._openai_plugin import openai_chat_service_builder as openai_chat_service_builder
from ._openai_plugin import openai_text_service_builder as openai_text_service_builder
from ._version import __version__ as __version__

__all__ = [name for name, obj in globals().items() if not (name.startswith("_") or inspect.ismodule(obj))]

del inspect



""":mod:`bodhiext.data_loader` bodhiext package for data loaders."""
import inspect

from ._file import FileLoader as FileLoader
from ._file import file_loader_service_builder as file_loader_service_builder
from ._plugin import bodhilib_list_services as bodhilib_list_services
from ._version import __version__ as __version__

__all__ = [name for name, obj in globals().items() if not (name.startswith("_") or inspect.ismodule(obj))]

del inspect



"""Cohere is a library for creating and managing a corpus of text documents."""
import inspect

from ._cohere_llm import Cohere as Cohere
from ._cohere_plugin import bodhilib_list_services as bodhilib_list_services
from ._cohere_plugin import cohere_llm_service_builder as cohere_llm_service_builder
from ._version import __version__ as __version__

__all__ = [name for name, obj in globals().items() if not (name.startswith("_") or inspect.ismodule(obj))]

del inspect



"""Plugin code for OpenAI services."""
from __future__ import annotations

import os
from typing import Any, Dict, List, Optional

from bodhilib import Service, service_provider

import openai

from ._openai_llm import OpenAIChat, OpenAIText
from ._version import __version__


@service_provider
def bodhilib_list_services() -> List[Service]:
    """Return a list of services supported by the plugin."""
    return [
        Service(
            service_name="openai_chat",
            service_type="llm",
            publisher="bodhiext",
            service_builder=openai_chat_service_builder,
            version=__version__,
        ),
        Service(
            service_name="openai_text",
            service_type="llm",
            publisher="bodhiext",
            service_builder=openai_text_service_builder,
            version=__version__,
        ),
    ]


def openai_text_service_builder(
    *,
    service_name: Optional[str] = None,
    service_type: Optional[str] = "llm",
    model: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> OpenAIText:
    """Returns an instance of OpenAIText LLM for the given arguments.

    Args:
        service_name (Optional[str]): service name to wrap, should be "openai_text"
        service_type (Optional[str]): service of the implementation, should be "llm"
        model (Optional[str]): OpenAI model identifier, e.g. text-ada-002
        api_key (Optional[str]): OpenAI api key, if not set, it will be read from environment variable OPENAI_API_KEY
        **kwargs: additional pass through arguments for OpenAI API client

    Returns:
        OpenAIText: an instance of OpenAIText implementing :class:`bodhilib.LLM`

    Raises:
        ValueError: if service_name is not "openai_text" or service_type is not "llm"
        ValueError: if api_key is not set and environment variable OPENAI_API_KEY is not set
    """
    if service_name != "openai_text" or service_type != "llm":
        raise ValueError(
            f"Unknown params: {service_name=}, {service_type=}, supported params: service_name='openai_text',"
            " service_type='llm'"
        )
    _set_openai_api_key(api_key)
    all_args: Dict[str, Any] = {"model": model, "api_key": api_key, **kwargs}
    all_args = {k: v for k, v in all_args.items() if v is not None}
    return OpenAIText(**all_args)


def openai_chat_service_builder(
    *,
    service_name: Optional[str] = None,
    service_type: Optional[str] = "llm",
    model: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> OpenAIChat:
    """Returns an instance of OpenAIChat LLM for the given arguments.

    Args:
        service_name: service name to wrap, should be "openai_chat"
        service_type: service type of the component, should be "llm"
        model: OpenAI chat model identifier, e.g. gpt-3.5-turbo
        api_key: OpenAI api key, if not set, it will be read from environment variable OPENAI_API_KEY
        **kwargs: additional arguments passed to the OpenAI API client

    Returns:
        OpenAIChat: an instance of OpenAIChat implementing :class:`bodhilib.LLM`

    Raises:
        ValueError: if service_name is not "openai" or service_type is not "llm"
        ValueError: if api_key is not set and environment variable OPENAI_API_KEY is not set
    """
    if service_name != "openai_chat" or service_type != "llm":
        raise ValueError(
            f"Unknown params: {service_name=}, {service_type=}, supported params: service_name='openai_chat',"
            " service_type='llm'"
        )
    _set_openai_api_key(api_key)
    all_args: Dict[str, Any] = {"model": model, "api_key": api_key, **kwargs}
    all_args = {k: v for k, v in all_args.items() if v is not None}
    return OpenAIChat(**all_args)


def _set_openai_api_key(api_key: Optional[str]) -> None:
    if api_key is None:
        if os.environ.get("OPENAI_API_KEY") is None:
            raise ValueError("environment variable OPENAI_API_KEY is not set")
        else:
            openai.api_key = os.environ["OPENAI_API_KEY"]
    else:
        openai.api_key = api_key



""":mod:`bodhiext.qdrant` module defines classes and methods for Qdrant Vector Database related operations."""
import uuid
from typing import Any, Dict, List, Optional, Union

from bodhilib import Distance, Embedding, Node, Service, VectorDB, VectorDBError, service_provider

# TODO: Once the module is separate, import this from the extension version file
from bodhilib import __version__ as bodhilib_version
from qdrant_client import QdrantClient
from qdrant_client.http.models import Distance as QdrantDistance
from qdrant_client.http.models import Filter, PointStruct, ScoredPoint, VectorParams

__version__ = bodhilib_version

_qdrant_distance_mapping = {
    Distance.COSINE.value: QdrantDistance.COSINE,
    Distance.DOT_PRODUCT.value: QdrantDistance.DOT,
    Distance.EUCLIDEAN.value: QdrantDistance.EUCLID,
}


class Qdrant(VectorDB):
    """Qdrant wraps the QdrantClient to provide a VectorDB interface."""

    def __init__(
        self,
        *,
        client: Optional[QdrantClient] = None,
        location: Optional[str] = None,
        url: Optional[str] = None,
        port: Optional[int] = 6333,
        grpc_port: int = 6334,
        prefer_grpc: Optional[bool] = False,
        https: Optional[bool] = None,
        api_key: Optional[str] = None,
        prefix: Optional[str] = None,
        timeout: Optional[float] = None,
        host: Optional[str] = None,
        path: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        """Mimics the QdrantClient interface.

        Raises:
            VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """
        if client:
            self.client = client
            return
        try:
            args = {
                "location": location,
                "url": url,
                "port": port,
                "grpc_port": grpc_port,
                "prefer_grpc": prefer_grpc,
                "https": https,
                "api_key": api_key,
                "prefix": prefix,
                "timeout": timeout,
                "host": host,
                "path": path,
                **kwargs,
            }
            args = {key: value for key, value in args.items() if value is not None}
            self.client = QdrantClient(**args)
        except (ValueError, RuntimeError) as e:
            raise VectorDBError(e) from e

    def ping(self) -> bool:
        return True

    def connect(self) -> bool:
        return True

    def close(self) -> bool:
        try:
            self.client.close()
            return True
        except (RuntimeError, ValueError) as e:
            raise VectorDBError(e) from e

    def get_collections(self) -> List[str]:
        try:
            collections = self.client.get_collections().collections
            names = [collection.name for collection in collections]
            return names
        except (ValueError, RuntimeError) as e:
            raise VectorDBError(e) from e

    def create_collection(
        self, collection_name: str, dimension: int, distance: Union[str, Distance], **kwargs: Dict[str, Any]
    ) -> bool:
        if not collection_name:
            raise VectorDBError(ValueError("Collection name cannot be empty"))
        if not dimension or dimension < 0:
            raise VectorDBError(ValueError("`dimension` cannot be empty"))
        if not distance:
            raise VectorDBError(ValueError("`distance` cannot be empty"))
        if str(distance) not in Distance.membersstr():
            raise VectorDBError(ValueError(f"Invalid distance: {distance}, valid values are {Distance.membersstr()}"))
        qdrant_distance = _qdrant_distance_mapping[str(distance)]
        params = {
            key: value for key, value in kwargs.items() if key in ["hnsw_config", "quantization_config", "on_disk"]
        }
        other_params = {
            key: value for key, value in kwargs.items() if key not in ["hnsw_config", "quantization_config", "on_disk"]
        }
        vector_config = VectorParams(size=dimension, distance=qdrant_distance, **params)
        try:
            result: bool = self.client.create_collection(collection_name, vectors_config=vector_config, **other_params)
            return result
        except (ValueError, RuntimeError) as e:
            raise VectorDBError(e) from e

    def delete_collection(self, collection_name: str, **kwargs: Dict[str, Any]) -> bool:
        try:
            result: bool = self.client.delete_collection(collection_name, **kwargs)
            return result
        except (ValueError, RuntimeError) as e:
            raise VectorDBError(e) from e

    def upsert(self, collection_name: str, nodes: List[Node]) -> List[Node]:
        try:
            for node in nodes:
                if node.id is None:
                    node.id = str(uuid.uuid4())

            points: List[PointStruct] = [
                PointStruct(id=node.id, vector=node.embedding, payload={"text": node.text, **node.metadata})
                for node in nodes
            ]
            _ = self.client.upsert(collection_name, points=points)
            return nodes
        except (ValueError, RuntimeError) as e:
            raise VectorDBError(e) from e

    def query(
        self,
        collection_name: str,
        embedding: Embedding,
        filter: Optional[Dict[str, Any]] = None,
        **kwargs: Dict[str, Any],
    ) -> List[Node]:
        try:
            if filter:
                qdrant_filter = _mongodb_to_qdrant_filter(filter)
                query_filter = Filter(**qdrant_filter)
            else:
                query_filter = None
            results = self.client.search(collection_name, embedding, query_filter=query_filter, **kwargs)
            return _to_nodes(results)
        except (ValueError, RuntimeError) as e:
            raise VectorDBError(e) from e


def qdrant_service_builder(
    *,
    service_name: Optional[str] = None,
    service_type: Optional[str] = None,
    publisher: Optional[str] = None,  # QdrantClient fails if passed extra args
    version: Optional[str] = None,  # QdrantClient fails if passed extra args
    client: Optional[QdrantClient] = None,
    url: Optional[str] = None,
    api_key: Optional[str] = None,
    timeout: Optional[float] = None,
    **kwargs: Dict[str, Any],
) -> Qdrant:
    """Returns an QDrant client instance.

    Lists only most used args required by Qdrant DB.
    """
    if service_name != "qdrant":
        raise ValueError(f"Unknown service: {service_name=}")
    if service_type != "vector_db":
        raise ValueError(f"Service type not supported: {service_type=}, supported service types: 'vector_db'")
    service_args: Dict[str, Any] = {
        "client": client,
        "url": url,
        "api_key": api_key,
        "timeout": timeout,
        **kwargs,
    }
    # TODO filter invalid args, only pass args recognized by QdrantClient
    return Qdrant(**service_args)


@service_provider
def bodhilib_list_services() -> List[Service]:
    """Returns a list of services supported by the plugin.

    Current supports vector_db service.
    """
    return [
        Service(
            service_name="qdrant",
            service_type="vector_db",
            publisher="bodhiext",
            service_builder=qdrant_service_builder,
            version=__version__,
        )
    ]


def _match(value: Any) -> Dict[str, Any]:
    if isinstance(value, (bool, int)):
        return {"value": value}
    if isinstance(value, str):
        return {"text": value}
    raise ValueError(f"Invalid type: {type(value)}")


def _mongodb_to_qdrant_filter(mongo_filter: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    qdrant_filter: Dict[str, Any] = {}
    if not mongo_filter:
        return qdrant_filter
    must_conditions: List[Dict[str, Any]] = []
    should_conditions: List[Dict[str, Any]] = []
    must_not_conditions: List[Dict[str, Any]] = []

    for key, value in mongo_filter.items():
        # Direct equal condition
        if isinstance(value, (bool, int, str)):
            must_conditions.append({"key": key, "match": _match(value)})

        # Special MongoDB conditions
        elif isinstance(value, dict):
            condition: Dict[str, Any] = {"key": key}
            if "$eq" in value:
                condition["match"] = _match(value["$eq"])
                must_conditions.append(condition)
            elif "$ne" in value:
                condition["match"] = _match(value["$ne"])
                must_not_conditions.append(condition)
            elif "$in" in value:
                for v in value["$in"]:
                    should_conditions.append({"key": key, "match": _match(v)})
            elif "$nin" in value:
                for v in value["$nin"]:
                    must_not_conditions.append({"key": key, "match": _match(v)})
            else:
                range_condition = {}
                if "$gt" in value:
                    range_condition["gt"] = value["$gt"]
                if "$gte" in value:
                    range_condition["gte"] = value["$gte"]
                if "$lt" in value:
                    range_condition["lt"] = value["$lt"]
                if "$lte" in value:
                    range_condition["lte"] = value["$lte"]
                if range_condition:
                    condition["range"] = range_condition
                    must_conditions.append(condition)

    if must_conditions:
        qdrant_filter["must"] = must_conditions
    if should_conditions:
        qdrant_filter["should"] = should_conditions
    if must_not_conditions:
        qdrant_filter["must_not"] = must_not_conditions

    return qdrant_filter


def _to_nodes(results: List[ScoredPoint]) -> List[Node]:
    nodes: List[Node] = []
    for result in results:
        text = result.payload.pop("text", "")
        node = Node(
            id=result.id,
            text=text,
            embedding=result.vector,
            metadata=result.payload,
        )
        nodes.append(node)
    return nodes



"""module for file data loader plugin for bodhilib."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional

from bodhilib import DataLoader, Document, PathLike
from bodhilib.logging import logger

LoaderCallable = Callable[[Path], List[Document]]

FILE_LOADERS: Dict[str, LoaderCallable] = {
    ".txt": lambda path: _txt_loader(path),
}


def _txt_loader(path: Path) -> List[Document]:
    return [Document(text=path.read_text(), metadata={"filename": path.name, "dirname": str(path.parent)})]


class FileLoader(DataLoader):
    """File data loader plugin for bodhilib.

    Supported file types:
        ".txt": reads txt file and returns a Document with text and metadata
    """

    def __init__(self) -> None:
        self.paths: List[Path] = []

    def add_resource(  # type: ignore
        self,
        *,
        files: Optional[List[PathLike]] = None,
        file: Optional[PathLike] = None,
        dir: Optional[PathLike] = None,
        recursive: bool = False,
    ) -> None:
        """Add a file or directory resource to the data loader with given :data:`~bodhilib.PathLike` location.

        Args:
            files (Optional[List[PathLike]]): A list of file paths to add.
            file (Optional[PathLike]): A file path to add.
            dir (Optional[PathLike]): A directory path to add.
            recursive (bool): Whether to add files recursively from the directory.

        Raises:
            ValueError: if any of the files or the dir provided does not exists.
        """
        if file:
            self._add_path(file)
        elif files:
            for path in files:
                self._add_path(path)
        elif dir:
            self._add_dir(dir, recursive)
        else:
            logger.info("paths or path must be provided")

    def __iter__(self) -> Iterator[Document]:
        """Return an iterator over the documents in the data loader."""
        return _FileIterator(self.paths)

    def _add_path(self, path: PathLike) -> None:
        if isinstance(path, str):
            path = Path(path)
        if not path.exists():
            raise ValueError(f"Path {path.absolute()} does not exist")
        self.paths.append(path)

    def _add_dir(self, dir: PathLike, recursive: bool) -> None:
        if isinstance(dir, str):
            dir = Path(dir)
        if not os.path.isdir(dir):
            raise ValueError(f"passed argument {dir=} is not a directory")
        if recursive:
            for root, _, files in os.walk(dir):
                for file in files:
                    self._add_path(os.path.join(root, file))
        else:
            for file in os.listdir(dir):
                self._add_path(os.path.join(dir, file))


class _FileIterator(Iterator[Document]):
    def __init__(self, paths: List[Path]) -> None:
        self.paths = paths
        self._generator = self._doc_generator()

    def __iter__(self) -> Iterator[Document]:
        return self

    def __next__(self) -> Document:
        return next(self._generator)

    def _doc_generator(self) -> Iterator[Document]:
        for path in self.paths:
            if path.suffix in FILE_LOADERS:
                yield from FILE_LOADERS[path.suffix](path)
            else:
                logger.warning(f"For filename={path}, file type {path.suffix} not supported, skipping")


def file_loader_service_builder(
    *,
    service_name: Optional[str] = "file",
    service_type: Optional[str] = "data_loader",
    publisher: Optional[str] = "bodhiext",
    **kwargs: Dict[str, Any],
) -> FileLoader:
    """Return a file data loader service builder.

    Builds and returns an instance of :class:`~bodhiext.data_loader._file.FileLoader`
    with the passed arguments.

    Returns:
        FileLoader (:class:`~bodhiext.data_loader._file.FileLoader`): FileLoader instance to load local files
            as resources.
    """
    if service_name != "file":
        raise ValueError(f"Unknown service: {service_name=}")
    if service_type != "data_loader":
        raise ValueError(f"Service type not supported: {service_type=}, supported service types: data_loader")
    if publisher is not None and publisher != "bodhiext":
        raise ValueError(f"Unknown publisher: {publisher=}")
    return FileLoader()



import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from bodhilib import BasePromptSource, Service, parse_prompt_template, service_provider
from bodhilib._models import PromptTemplate

from ._version import __version__

CURRENT_DIR = Path(os.path.dirname(os.path.abspath(__file__)))
DEFAULT_TEMPLATES_DIR = CURRENT_DIR / "data"


class LocalDirectoryPromptSource(BasePromptSource):
    """BodhiPromptSource is a prompt source implementation by bodhiext."""

    def __init__(self, source_dir: Optional[str] = None) -> None:
        """Initializes BodhiPromptSource.

        Args:
            source_dir (Optional[str]): Path to the directory containing prompt templates. Defaults to None.
                If None, defaults to bodhiext's default prompt template directory.
        """
        if source_dir is None:
            source_dir = str(DEFAULT_TEMPLATES_DIR)
        self.source_dir = source_dir
        self.templates: Optional[List[PromptTemplate]] = None

    def _find(self, tags: List[str]) -> List[PromptTemplate]:
        if not self.templates:
            self.templates = self._load_templates()
        return [template for template in self.templates if set(tags).issubset(set(template.metadata.get("tags", [])))]

    def _list_all(self) -> List[PromptTemplate]:
        if not self.templates:
            self.templates = self._load_templates()
        return self.templates

    def _load_templates(self) -> List[PromptTemplate]:
        templates = []
        # recursively find all files in the source_dir and parse templates
        for root, _, files in os.walk(self.source_dir):
            for file in files:
                with open(os.path.join(root, file), "r") as f:
                    text = f.read()
                    parsed_templates = parse_prompt_template(text)
                    templates.extend(parsed_templates)
        return templates


@service_provider
def bodhilib_list_services() -> List[Service]:
    """Returns a list of services supported by the plugin.

    Currently supports prompt_source service.
    """
    return [
        Service(
            service_name="local_dir_prompt_source",
            service_type="prompt_source",
            publisher="bodhiext",
            service_builder=bodhi_prompt_source_builder,
            version=__version__,
        )
    ]


def bodhi_prompt_source_builder(
    *,
    service_name: Optional[str] = None,
    service_type: Optional[str] = None,
    publisher: Optional[str] = None,  # QdrantClient fails if passed extra args
    version: Optional[str] = None,  # QdrantClient fails if passed extra args
    **kwargs: Dict[str, Any],
) -> LocalDirectoryPromptSource:
    """Returns an instance of BodhiPromptSource."""
    if service_name != "local_dir_prompt_source" or service_type != "prompt_source" or publisher != "bodhiext":
        raise ValueError(
            f"Invalid arguments to the service builder: {service_name=}, {service_type=}, {publisher=}, supported"
            " values are: service_name='local_dir_prompt_source',{service_type='prompt_source', publisher='bodhiext'"
        )

    return LocalDirectoryPromptSource()
